%% R = process(X, Y, N, K, T,alpha)
% predicts given data using polynomial models of various degree, as well as
% with various model selection methods, including the switch distribution.
%
% Input:
% X,Y: X and Y values of the data to be predicted
% N: the number of outcomes to process
% K: the number of models to use
% T: the type of switch distribution to be used:
%    - T=0: fast switch distribution; alpha determines freezing times
%    - T=1: slow switch distribution
%    - T=2: basic switch distribution with K_n = L_n
%
% The order k model contains all polynomials of degree k, with standard
% normal noise. The prediction strategies associated with the models are
% Bayesian, with Jeffreys' prior. The Bayesian posterior density is
% Gaussian, with mean at the least squares polynomial and sigma at some
% value larger than 1. Prediction starts at n=K+3.
%
% Return values:
% R: a structure with the following fields:
% R.coef{k}, R.A{k}: result of the order k fit for the entire sample
% R.M_switch(n):    model selected by switch distribution at sample size n
% R.M_bayes(n):     model selected by Bayes factors at sample size n
% R.M_aic(n):       model selected by AIC at sample size n (only if T=0)
% R.M_bic(n):       model selected by BIC at sample size n (only if T=0)
% R.L(k,1:n):       log(P_k(x^n))
% R.L_switch(n):    log(P_switch(x^n))
% R.L_bayes:        log(P_BMA(x^n))
% R.L_aic:          log(P_AIC(x^n)) (only if T=0)
% R.L_bic:          log(P_BIC(x^n)) (only if T=0)
%
function R = process(X, Y, N, K, T, alpha)
  logtheta = log(0.5);
  log1mtheta = log(0.5); % theta is prob. of switch not being last
  logK    = log(K);

  log_alpha = log(alpha);
  block_end = 0; % (for T=2) make sure to start a new block immediately
    
  R.L = zeros(K,N);
  Lw  = repmat(-Inf, 1, K); % initialise L weights
  
  if T==0, R.L_aic = zeros(1,N); R.L_bic = zeros(1,N); end  
  if T==1, Kw = []; Ksize = 0; else Kw = repmat(-Inf, 1, K); Ksize = K; end

  %% Fit results
  SSE       = [];
  Lcoef     = {};
  LA        = {};

  logpool = 0;                  % initial probability mass 1

  for n=K+1:N
    %% Update all prediction strategies in L with X^{n-1}, Y^{n-1}
    % store the logprobabilities for the next outcome (X_n,Y_n) in lw

    for k = 1:K
      % instead of re-fitting each time, this could be sped up by tracking the
      % parameter estimate sequentially.
      [Lcoef{k}, LA{k}, SSE(k)] = mypolyfit(X(1:n-1), Y(1:n-1), k-1);
      [mu, s2] = mypolyval(Lcoef{k}, LA{k}, X(n));
      lw(k) = lgauss(Y(n), mu, s2);
      R.L(k,n) = R.L(k,n-1) + lw(k);
    end
        
    % Update all prediction strategies in K and store the logprobabilities
    % for the next outcome (X_n,Y_n) in kw
    switch T
      case 0
        %% fast switch distribution
        if n > block_end
          kw = lw;
          block_start = n+1;
          block_end   = min(N, floor(alpha.^ceil(log(n+1)/log_alpha)));
          for k=1:K   
            [mu, s2]   = mypolyval(Lcoef{k}, LA{k}, X(block_start:block_end));
            Kbuffer{k} = lgauss(Y(block_start:block_end), mu, s2);
          end
        else
          for k=1:K, kw(k) = Kbuffer{k}(n-block_start+1); end
        end
      case 1
        %% slow switch distribution
        for k=1:K
          [mu, s2] = mypolyval(Lcoef{k}, LA{k}, X(n:N));
          Kbuffer{Ksize + k} = lgauss(Y(n:N), mu, s2);
        end
        Kw = [Kw, repmat(-Inf, 1, K)];
        Ksize = Ksize + K;
        for i=1:Ksize
          n0 = floor((i-1)/K)+K+1; % index of first outcome in Kbuffer{i}
          kw(i) = Kbuffer{i}(n-n0+1);
        end
      case 2
        %% "simple" switch distribution with K=L
        kw = lw;
    end

    %% Distribute the pool over the updated K and L
    % Use theta=1/2 and uniform lambda_t and kappa_t
    wkd = logpool + logtheta   - log(Ksize);
    wld = logpool + log1mtheta - logK;
    Kw  = logsum([Kw ; repmat(wkd, 1, Ksize) ]); 
    Lw  = logsum([Lw ; repmat(wld, 1, K)     ]);

    %% Predict Y(n) given X(n) and incur logarithmic loss
    Lw = Lw + lw;
    Kw = Kw + kw;
	
    %% Calculate AIC and BIC
    if T==0
      ML = 0.5*SSE; % + 0.5*n*log(2*pi), but this constant cancels
      AIC = ML + (1:K);
      BIC = ML + 0.5.*(1:K).*log(n);
      [dummy, R.M_aic(n)] = min(AIC);
      [dummy, R.M_bic(n)] = min(BIC);
      R.L_aic(n) = R.L_aic(n-1) + logsum((lw-AIC)')-logsum(-AIC');
      R.L_bic(n) = R.L_bic(n-1) + logsum((lw-BIC)')-logsum(-BIC');
    end

    %% Gather probability mass from K into the pool
    pswitch       = 1./(n+1);
    logpswitch    = log(pswitch);
    logpnoswitch  = log(1-pswitch);
    logsumKw      = logsum(Kw');
    logpool       = logsumKw + logpswitch;
    Kw            = Kw + logpnoswitch;
    R.L_switch(n) = logsum([logsumKw, Lw]');

    %% Record MAP coefficients and sigma (ignores freezing time)
    [dummy, R.M_switch(n)] = max(Lw);
  end
  R.coef = Lcoef;
  R.A    = LA;
  R.L_bayes = logsum(R.L) - logK;
  [dummy, R.M_bayes] = max(R.L);

end
