%% visualise(D, R)
% Visualise data D with experimental results R.
%
% Input:
% D: data set, possibly containing multiple runs
% R: experimental results obtained with R = gather(D, ...)
%
% Output:
% Nice graphs
%
function visualise(D, R)
  assert(R.types(1)==1);

  T = D.T - repmat(D.T(:,R.K+1), 1, size(D.T,2));
  T(:,1:R.K) = 0;

  %% Plot the data, the trend and the model selected by the switch distribution
  figure(1); clf;
  hold on;
  set(gca, 'FontSize', 16);
  plot(D.X(:,1), D.Y(:,1), '.b');
  s = (-1:0.01:1)';
  y0 = min(D.f(s)); y1 = max(D.f(s));
  axis([-1,1,y0-1.8, y1+1.8]);
  plot(s, D.f(s), 'k', 'linewidth', 2);
  k1 = R.M_bayes(1,end);
  k2 = R.M_sw0(1,end);
  if k2 ~= k1
    plot(s, mypolyval(R.coef{k1}, R.A{k1}, s), 'g', 'linewidth', 2);
  end
  plot(s, mypolyval(R.coef{k2}, R.A{k2}, s), 'r', 'linewidth', 2);
  L = {'Data', 'Trend', sprintf('Degree %d', k1-1)};
  if k2 ~= k1, L = [ L, { sprintf('Degree %d', k2-1) } ]; end
  legend(L, 'Location', 'Best');

  %% Plot the MAP degrees for BMA and Switch
  figure(2); clf;
  hold on;
  grid on;
  set(gca, 'FontSize', 16);
  
  L = {'AIC', 'BIC', 'Bayes', 'Fast switch'};
  O = {R.M_aic, R.M_bic, R.M_bayes, R.M_sw0};
  if R.types(2), L = [ L, { 'Slow switch' } ]; O = [O, {R.M_sw1}]; end
  if R.types(3), L = [ L, { 'Default Switch' } ]; O = [O, {R.M_sw2}]; end
  
  y1 = max(mean(O{1}));
  
  step = 1;
  if y1 > 10, step = 5; end
  if y1 > 50, step = 10; end
  set(gca, 'YTick', [0:step:y1]);
  
  for i=1:length(L)
    mo = max(mean(O{i}));
    if mo > y1, y1=mo; end
    fprintf('Max order for %-12s: %3d\n', L{i}, max(max(O{i})));
  end

  axis([R.K+2, D.N, 0, (y1-1)*1.1]);
  
  plot(mean(R.M_aic)-1, 'b', 'LineWidth', 2);
  plot(mean(R.M_bic)-1, 'c', 'LineWidth', 2);
  plot(mean(R.M_bayes)-1, 'g', 'LineWidth', 2);
  plot(mean(R.M_sw0)-1, 'r', 'LineWidth', 2);
  if R.types(2), plot(mean(R.M_sw1)-1, 'm', 'LineWidth', 2); end
  if R.types(3), plot(mean(R.M_sw2)-1, 'k', 'LineWidth', 2); end
  legend(L, 'Location', 'SouthEast');
  
  %% Plot the cumulative logloss compared to the true density
  figure(3); clf;
  hold on;
  set(gca, 'FontSize', 16);

  trans = @(X)mean(T(:,1:length(X))-X);
  y1 = max(max([trans(R.L_aic); trans(R.L_bic); trans(R.L_bayes); trans(R.L_sw0)]));
  axis([R.K, D.N, 0, y1*1.05./log(2)]);
  
  plotrisk(T, R.L_aic,    'b');
  plotrisk(T, R.L_bic,    'c');
  plotrisk(T, R.L_bayes,  'g');
  plotrisk(T, R.L_sw0, 'r');
  
  % errorbar(trans(R.L_sw0), std(R.L_sw0)./100, 'xr');
  if R.types(2), plotrisk(T, R.L_sw1, 'm'); end
  if R.types(3), plotrisk(T, R.L_sw2, 'k'); end
  legend(L, 'Location', 'SouthEast');
end


%% plotrisk(T, L, style)
% Plot the risk in the desired style.
%
% Input:
% T(n): cumulative log of the true density for the first n outcomes
% L(n): cumulative log of predicted density for the first n outcomes
% style: plot style, supplied directly to the plot command
%
function plotrisk(T, L, style)
  [runs, N] = size(L);
  R = (T(:,1:N)-L)./log(2); % convert to bits
  plot(mean(R), style, 'LineWidth', 2);
  X = 1:floor(N./20):N;
  R2 = R(:,X);
  errorbar(X,mean(R2),std(R2)./sqrt(runs), ['+',style], 'HandleVisibility', 'off');
end
