%% Markov
% Implements simple Markov model for binary sequences.
%
classdef Markov < PFS
    properties
        order;        % Markov chain order
        nctxt;        % number of contexts
        counts;       % occurrence counts for all symbols in all contexts
        sumcounts;    % occurrence counts for all contexts (sum of counts)
        context = 0;  % the current context as an integer
    end
    
    methods

        %% Markov(order)
        % Constructs a Markov model of the given order
        function this = Markov(order)
            this.order     = order;
            this.nctxt     = 128^order;
            this.counts    = zeros(128, this.nctxt);
            this.sumcounts = zeros(this.nctxt);
        end
        
        %% clone() [required by PFS]
        function PFS = clone(this)
            PFS = Markov(this.order);
            PFS.counts    = this.counts;
            PFS.context   = this.context;
            PFS.nctxt     = this.nctxt;
            PFS.sumcounts = this.sumcounts;
        end
        
        %% predict(x) [required by PFS]
        % Uses the Krichevsky-Trofimov estimator to calculate the
        % probability of observing x in the current context.
        function p = predict(this, x)
            c = this.context;
            s = 0.5; % number of virtual initial observations for each symbol
            p = (this.counts(double(x)+1,c+1)+s)/(this.sumcounts(c+1)+s*128);
        end
        
        %% process(x) [required by PFS]
        % Updates counts, sumcounts and context with the given symbol.
        function process(this, x)
            c = this.context;
            i = double(x);
            this.counts(i+1, c+1) = this.counts(i+1, c+1) + 1;
            this.sumcounts(c+1) = this.sumcounts(c+1) + 1;
            this.context = mod(128*c+i, this.nctxt);
        end
        
    end
end