%% PFS_mixture
% Interface for mixtures of PFSs, itself a PFS. It manages the set of
% prediction strategies that are mixed over.
%
classdef PFS_mixture < PFS

    properties (GetAccess = protected)
        core     % the core maintains all the weights
        PFSs     % the set of PFSs to mix over
        Nactive; % the number of PFSs that have not been frozen
    end
    
    methods
        
        %% PFS_mixture(PFSs, core)
        % Constructs a PFS_mixture with the specified PFSs and core
        function this = PFS_mixture(PFSs, core, Nactive)
            this.core    = core;         
            this.PFSs    = PFS_mixture.clones(PFSs);
            this.Nactive = Nactive;
        end
        
        %% clone() [required by PFS]
        % Implement this method to be able to use the slow/fast switch
        % distribution with mixture PFSs.
        function PFS = clone(this)
            error('Cloning is not implemented for mixtures');
        end
        
        %% predict(x) [required by PFS]
        function p = predict(this, x)
            probs = cellfun(@(e)e.predict(x), this.PFSs);
            p = dot(probs, this.core.weights);
        end
        
        %% process(x) [required by PFS]
        function process(this, x)
            probs = cellfun(@(e)e.predict(x), this.PFSs);
            this.core.update(probs);
            cellfun(@(e)e.process(x), this.PFSs(1:this.Nactive));
        end
    end
    
    methods (Static)
        %% clones(PFSs)
        % returns clones of the PFSs in the given cell array
        function clones = clones(PFSs)
            clones = cellfun(@(e)e.clone(), PFSs, 'UniformOutput', false);
        end
    end
    
    methods (Abstract)

        %% model_weights()
        % Implementing classes should be able to return model weights
        % appropriate for consistent model selection.
        model_weights(this)
        
    end
      
end
