%% core
% Manages a bunch of weights.
%
% Warning: problems may arise if any of the weights becomes so small that
% it cannot be properly represented by a machine precision floating point
% number, which is not necessarily an unrealistic scenario.
% The solution is to store all the weights to logarithmic precision:
% Given two numbers x=log(a) and y=log(b) with a>b, we can calculate
% log(a+b) = log(a*(1+b/a)) = x + log(1+exp(y-x))), which is numerically safe.
% If the weights are stored this way, there is no longer any need for
% normalisation.
% For simplicity, this code does not deal with such issues.
%
classdef core < handle
    
    properties (SetAccess = protected)
        weights;
    end
    
    methods

        %% core(weights)
        % Constructs a core with the given initial weights
        function this = core(weights); this.weights = weights; end
        
        %% update(p)
        % Updates the weights for all prediction strategies based on the
        % probabilities they assigned to the observed outcome.
        function update(this, p)
            % extend weight vector if necessary
            if size(p,2) > size(this.weights,2)
                this.weights(end+1:size(p,2)) = 0;
            end
            
            % update weights and renormalize
            w = this.weights .* p;
            this.weights = w / sum(w);
        end
    end
        
end
