%% switch_fast
% Implementation of the fast switch distribution. The constituent PFSs are
% frozen and refrozen at exponential intervals.
%
classdef switch_fast < PFS_mixture
    
    properties (GetAccess = private)
        block_end; % index where we refreeze all switching experts
        alpha;     % parameter that determines when freezing occurs
    end
    
    methods

        %% switch_fast(PFSs, alpha)
        % Constructs a switch_fast PFS over the given set of PFSs.
        function this = switch_fast(PFSs, alpha)
            Kst = length(PFSs);
            this = this@PFS_mixture([PFSs PFSs], switch_core(Kst, Kst), Kst);
            this.alpha = alpha;
            this.refreeze();
        end
        
        %% process(x)
        % Refreezes the PFSs when necessary.
        function process(this, x)
            this.process@PFS_mixture(x);
            if this.core.N >= this.block_end, this.refreeze(); end
        end
        
        %% model_weights(x) [required by PFS_mixture]
        % Conditions the weights on only the set of stable PFSs. This
        % is the model selection criterion defined in the paper.
        function w = model_weights(this)
            w = this.core.weights(1:this.core.Kst);
            w = w/sum(w);
        end
        
        
    end
    
    methods (Access = private)
        
        %% refreeze()
        % Brings the predictions of all frozen PFSs up to date.
        % Block i starts at index floor(a^i). But if a is small, we may
        % need to skip some of the blocks. We solve this by calculating
        % the smallest i such that the current sample falls within
        % block i-1. We must have floor(a^i)>n+1, or block i would
        % already have ended, so we require i>=log(n+1)/log(a). The
        % smallest such i is i=ceil(log(b)/log(a)).
        function refreeze(this)
            Kst = this.core.Kst;
            this.PFSs(Kst+1:end) = PFS_mixture.clones(this.PFSs(1:Kst));
            a = this.alpha;
            n = this.core.N;
            fprintf('switch_fast: refreezing at n=%d\n', n);
            this.block_end = floor(a ^ ceil(log(n+1)/log(a))); 
            assert(this.block_end >= n, 'Attempt to increase block size failed: went from %d to %d', n, this.block_end);
        end
        
    end

end